! Copyright (C) 2023 Aleksander Sabak.
! See https://factorcode.org/license.txt for BSD license.
USING: assocs help.markup help.markup.private help.syntax kernel sequences strings urls ;
IN: FALSE

HELP: <FALSE-state>
{ $values
    { "stack" sequence } { "dictionary" assoc }
    { "state" FALSE-state }
}
{ $description "Creates a new FALSE state with clones of " { $snippet "stack" } " and " { $snippet "dictionary" } " as the initial stack and initial variable dictionary respectively." }
{ $see-also <empty-FALSE-state> with-FALSE } ;

HELP: <empty-FALSE-state>
{ $values
    { "state" FALSE-state }
}
{ $description "Creates a new, empty FALSE state. It has no bound variables and an empty stack." }
{ $see-also <FALSE-state> with-FALSE } ;

HELP: FALSE-state
{ $class-description "The class of tuples holding states of FALSE execution to be operated on by compiled FALSE programs. Implements " { $link "sequence-protocol" } " for access to stack and " { $link "assocs-protocol" } " for access to variables." }
{ $see-also <FALSE-state> <empty-FALSE-state> } ;

HELP: FALSE[
{ $syntax "FALSE[ 0 0[ß^$$47>\\58\\>&][48-\\10*+]#%[$1>][$1-]#[\\$0=~][*]#%.]" }
{ $description "Syntax for a FALSE program. It will run on a " { $link FALSE-state } " object. Syntax and semantics of FALSE language are explained in " { $link "FALSE" } "." }
{ $errors "Throws an error when the unsupported " { $snippet "`" } " command is encountered." }
{ $examples
  { $example
    "USING: kernel accessors FALSE prettyprint ;"
    "{ 2 1 3 7 } { } <FALSE-state>
    FALSE[ +-*] call stack>> ."
    "V{ -18 }"
  }
  { $example
    "USING: FALSE prettyprint ;"
    "\"Hello\" FALSE[ \\1+] with-FALSE ."
    "\"Helom\""
  }
}
{ $see-also with-FALSE } ;

HELP: scan-FALSE
{ $values
    { "quote" { $quotation ( state -- state' ) } }
}
{ $description "Reads a FALSE program from parser input until a closing " { $snippet "[" } " and compiles into a quotation to run on a " { $link FALSE-state } "." }
{ $errors "Throws an error when the unsupported " { $snippet "`" } " command is encountered." }
{ $see-also POSTPONE: FALSE[ } ;

HELP: unsupported-FALSE-command
{ $values
    { "command" object }
}
{ $description "Throws an " { $link unsupported-FALSE-command } " error." }
{ $error-description "Thrown during FALSE compilation if the " { $snippet "`" } " command is encountered." } ;

HELP: with-FALSE
{ $values
    { "seq" sequence } { "q" { $quotation ( ..A state -- ..B state' ) } }
    { "seq'" sequence }
}
{ $description "Wrapper around quotations transforming FALSE state. Creates a new " { $link FALSE-state } " instance with " { $snippet "seq" } " as its initial stack, runs " { $snippet "q" } " on it and extracts the final stack into a new sequence of the same type as " { $snippet "seq" } "." }
{ $examples
  { $example
    "USING: FALSE prettyprint ;"
    "{ 2 1 3 7 } FALSE[ +-*] with-FALSE ."
    "{ -18 }"
  }
  { $example
    "USING: FALSE prettyprint ;"
    "\"Hello\" FALSE[ \\1+] with-FALSE ."
    "\"Helom\""
  }
}
{ $see-also <FALSE-state> <empty-FALSE-state> }
;

ARTICLE: "FALSE" "FALSE language"
{ { $url URL"https://esolangs.org/wiki/FALSE" "FALSE" } " is a stack-based esoteric language designed by " { $url URL"https://strlen.com/" "Wouter van Oortmerssen" } " in 1993, with the goal of creating a powerful and obfuscated language with as small a compiler as possible." }

{ $heading "Overview of FALSE" }
FALSE has a simple concatenative syntax where each character is one command, apart from numbers which can span multiple characters. It operates on a stack but it can also store values in 26 variables named { $snippet "a" } through { $snippet "z" } . It works with three types of values:
{ $list
  { "Signed integers, for which literals are strings of digits or " { $snippet "'" } " followed by a character" }
  { "Variable references, for which literals are their names" }
  { "Quotations, for which literals are programs enclosed in " { $snippet "[...]" } }
}
It provides a set of builtin operators:
{ $table 
  { { $snippet "$" } { $snippet ( x -- x x ) } "Duplicates top of the stack" }
  { { $snippet "%" } { $snippet ( x -- ) } "Drop top of the stack" }
  { { $snippet "\\" } { $snippet ( x y -- y x ) } "Swaps top two elements on the stack" }
  { { $snippet "@" } { $snippet ( x y z -- y z x ) } "Pulls the third element from the stack to the top" }
  { { $snippet "ø" } { $snippet ( x n*x n -- x n*x x ) } "Replaces a number from top of the stack with an element that deep into the stack" }
  { { $snippet "O" } { $snippet ( x n*x n -- x n*x x ) } "Nonstandard, same as ø" }
  { { $snippet "+" } { $snippet ( x y -- z ) } "Adds top two numbers on the stack" }
  { { $snippet "-" } { $snippet ( x y -- z ) } "Subtracts top from second on the stack" }
  { { $snippet "*" } { $snippet ( x y -- z ) } "Multiplies top two numbers on the stack" }
  { { $snippet "/" } { $snippet ( x y -- z ) } "Divides second on the stack by top" }
  { { $snippet "_" } { $snippet ( x -- y ) } "Negates top of the stack" }
  { { $snippet "&" } { $snippet ( x y -- z ) } { "Performs a bitwise " { $snippet "and" } " on top two values o the stack" } }
  { { $snippet "|" } { $snippet ( x y -- z ) } { "Performs a bitwise " { $snippet "or" } " on top two values o the stack" } }
  { { $snippet "~" } { $snippet ( x -- y ) } "Performs a bitwise negation on top of the stack" }
  { { $snippet ">" } { $snippet ( x y -- z ) } "Pushes -1 if second on the stack is greater than top, otherwise pushes 0" }
  { { $snippet "=" } { $snippet ( x y -- z ) } "Pushes -1 if top two on the stack are equal, otherwise pushes 0" }
  { { $snippet "!" } { $snippet ( quot -- ) } "Executes top of stack" }
  { { $snippet "?" } { $snippet ( ? quot -- ) } "Executes top of stack if second on stack is non-zero" }
  { { $snippet "#" } { $snippet ( pred body -- ) } { "Executes " { $snippet "pred" } ", pops its return value, and loops calling " { $snippet "body" } ", " { $snippet "pred" } " and popping until the value popped is zero" } }
  { { $snippet ";" } { $snippet ( ref -- x ) } "Fetch from a variable reference" }
  { { $snippet ";" } { $snippet ( x ref -- ) } "Store into a variable reference" }
  { { $snippet "^" } { $snippet ( -- x ) } "Fetch a character from input stream, or -1 if input stream is exhausted" }
  { { $snippet "," } { $snippet ( x -- ) } "Print a character with a given code to the output stream" }
  { { $snippet "." } { $snippet ( x -- ) } "Print a number to the output stream" }
  { { $snippet "ß" } { $snippet ( -- ) } "Flush the input/output stream" }
  { { $snippet "B" } { $snippet ( -- ) } "Nonstandard, same as ß" }
  { { $snippet "`" } { $snippet ( -- * ) } "Unsupported" }
}
It also has two additional syntax constuctions:
{ $list
  { { $snippet "\"...\"" } " - print string enclosed in quotation marks" }
  { { $snippet "{...}" } " - ignore anything in curly braces" }
}

{ $heading "Vocabulary" }
The { $vocab-link "FALSE" } vocabulary provides a FALSE to Factor compiler in two words:
{ $subsections scan-FALSE POSTPONE: FALSE[ }
These offer a way to parse FALSE into quotations that take and return a FALSE state object. State objects can be constructed from a sequence which will form the initial stack and an assoc which will define initial variable bindings. The vocabulary also includes a wrapper word for using a FALSE quotation as a function transforming a sequence:
{ $subsections FALSE-state with-FALSE }

{ $heading "Implementation specific features" }
This FALSE to Factor compiler differs from the standard in several ways:
{ $list
  { "It doesn't support the " { $snippet "`" } " command" }
  { "It supports " { $snippet "O" } " and " { $snippet "B" } " commands as aliases to " { $snippet "ø" } " and " { $snippet "ß" } " respectively" }
  { "Any character not recognised as a FALSE command can be used as a variable reference, and variable references are stored as numbers on the stack which allows any number to be a variable reference as well" }
}

;

ABOUT: "FALSE"
